package com.bosch.tt.de.emma.connector;

import android.arch.lifecycle.LifecycleOwner;
import android.arch.lifecycle.ProcessLifecycleOwner;
import android.os.Build;
import android.support.annotation.NonNull;
import android.support.annotation.Nullable;
import android.util.Log;

import com.bosch.sh.connector.thirdparty.ThirdPartyConnectorLibraryImpl;
import com.bosch.sh.connector.thirdparty.api.ThirdPartyConnectorLibrary;
import com.bosch.sh.connector.thirdparty.api.connection.ConnectionListener;
import com.bosch.sh.connector.thirdparty.api.http.Result;
import com.bosch.sh.connector.thirdparty.api.pairing.PairingListener;
import com.bosch.sh.connector.thirdparty.api.pairing.PairingResult;
import com.bosch.sh.connector.thirdparty.api.pairing.PairingStatus;
import com.bosch.sh.connector.thirdparty.api.pairing.PairingStatusListener;
import com.bosch.sh.connector.thirdparty.api.shc.ShcIdChangeResult;
import com.bosch.sh.connector.thirdparty.api.shc.ShcIdChangeResultListener;
import com.bosch.tt.de.emma.connector.service.RequestService;
import com.bosch.tt.de.emma.connector.service.ShcCallback;
import com.bosch.tt.de.emma.connector.utils.JsonConverter;
import com.facebook.react.bridge.Arguments;
import com.facebook.react.bridge.Callback;
import com.facebook.react.bridge.Promise;
import com.facebook.react.bridge.ReactApplicationContext;
import com.facebook.react.bridge.ReactContextBaseJavaModule;
import com.facebook.react.bridge.ReactMethod;
import com.facebook.react.bridge.WritableMap;
import com.facebook.react.modules.core.DeviceEventManagerModule;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.json.JSONTokener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class RNShcConnectorModule extends ReactContextBaseJavaModule implements ConnectionListener {

	private static final String TAG = "RNShcConnector";
	private static final Logger LOG = LoggerFactory.getLogger(RNShcConnectorModule.class.getSimpleName());

	private static final String RNSHC_CONNECTED = "RNSHC_CONNECTED";
	private static final String RNSHC_CONNECTION_LOST = "RNSHC_CONNECTION_LOST";
	private static final String RNSHC_DISCONNECTED = "RNSHC_DISCONNECTED";
	private static final String RNSHC_ID_CHANGE_RESULT = "RNSHC_ID_CHANGE_RESULT";
	private static final String RNSHC_PARING_STATUS_CHANGED = "RNSHC_PARING_STATUS_CHANGED";
	private static final String RNSHC_PARING_STATUS = "RNSHC_PARING_STATUS";

	private final ReactApplicationContext reactContext;

	/**
	 * The controller for the SHC connection.
	 */
	private ThirdPartyConnectorLibrary thirdPartyConnectorLibrary;


	public RNShcConnectorModule(ReactApplicationContext reactContext) {
		super(reactContext);
		this.reactContext = reactContext;
	}

	@Override
	public String getName() {
		return "RNShcConnector";
	}

	private synchronized ThirdPartyConnectorLibrary getThirdPartyConnectorLibrary() {
		if (thirdPartyConnectorLibrary == null) {
			Log.i(TAG, "Creating connection controller.");
			LifecycleOwner lifecycleOwner = ProcessLifecycleOwner.get();
			thirdPartyConnectorLibrary = ThirdPartyConnectorLibraryImpl.init(this.reactContext, lifecycleOwner);
		}

		return thirdPartyConnectorLibrary;
	}

	@ReactMethod
	public void initModule() {
		getThirdPartyConnectorLibrary();
	}

	@ReactMethod
	public void connect() {
		Log.v(TAG, "connect");

		getThirdPartyConnectorLibrary().connectionManager().connect(this);
	}

	@Override
	public void connectionLost(Exception e) {
		Log.i(TAG, "onShcConnectionLost");

		WritableMap params = Arguments.createMap();

		if (e != null) {
			params.putString("error", e.getMessage());
			Log.e(TAG, params.getString("error"));
		}

		sendEvent(RNSHC_CONNECTION_LOST, params);
	}

	@Override
	public void connected() {
		Log.i(TAG, "onShcConnected");

		sendEvent(RNSHC_CONNECTED, null);
	}

	@Override
	public void disconnected() {
		Log.i(TAG, "onShcDisconnected");

		sendEvent(RNSHC_DISCONNECTED, null);
	}

	@ReactMethod
	public void hasShcIdBeenProvided(Promise promise) {
		Log.v(TAG, "hasShcIdBeenProvided");
		if (getThirdPartyConnectorLibrary().hasShcIdBeenProvided()) {
			promise.resolve("The SHC ID has already provided");
		} else {
			promise.reject("SHC_ID_NOT_SET");
		}
	}

	@ReactMethod
	public void setShcId(final @NonNull String shcId, final Promise promise) {
		Log.v(TAG, "setShcId " + shcId);

		getThirdPartyConnectorLibrary().setShcId(shcId, new ShcIdChangeResultListener() {
			@Override
			public void onShcIdSet(ShcIdChangeResult shcIdChangeResult) {
				RNShcConnectorModule.this.sendShcIdChangedEvent(shcIdChangeResult);
			}
		});
	}

	@ReactMethod
	public void setShcIdAndLocalAddress(final @NonNull String shcId, final @NonNull String address, final Promise promise) {
		Log.v(TAG, "setShcIdAndLocalAddress: " + shcId + ", " + address);

		getThirdPartyConnectorLibrary().setShcId(shcId, new ShcIdChangeResultListener() {
			@Override
			public void onShcIdSet(ShcIdChangeResult shcIdChangeResult){
				RNShcConnectorModule.this.sendShcIdChangedEvent(shcIdChangeResult);
				getThirdPartyConnectorLibrary().setLocalIp(address);
				promise.resolve(null);
			}
		});

	}

	@ReactMethod
	public void checkPairStatus(Promise promise) {
		Log.v(TAG, "Get paired Status ");
		thirdPartyConnectorLibrary.checkPairingStatus(new PairingStatusListener() {
			@Override
			public void receivedPairingStatus(PairingStatus pairingStatus) {
				Log.v(TAG, "Check pair Status " + pairingStatus.name());
				WritableMap params = Arguments.createMap();
				params.putString("status", pairingStatus.name());
				sendEvent(RNSHC_PARING_STATUS, params);
			}
		});
	}

	@ReactMethod
	public void pair(final String id, final String nonce, final Promise promise) {
		thirdPartyConnectorLibrary.pair(id, nonce, "EM_" + Build.MODEL, new PairingListener() {
			@Override
			public void pairingFinished(PairingResult pairingResult) {
				Log.v(TAG, "Pair Status Event " + pairingResult.name());
				WritableMap params = Arguments.createMap();
				params.putString("status", pairingResult.name());
				sendEvent(RNSHC_PARING_STATUS_CHANGED, params);
			}
		});
	}

	@ReactMethod
	public void get(final String url, final Callback successCallback, final Callback errorCallback) {
		RequestService.get(url, getThirdPartyConnectorLibrary(), new ShcCallback() {
			@Override
			public void onResult(Result result) {
				handleResponse(result, successCallback, errorCallback);
			}
		});
	}

	@ReactMethod
	public void put(final String url, final String body, final Callback successCallback, final Callback errorCallback) {
		RequestService.put(url, body, getThirdPartyConnectorLibrary(), new ShcCallback() {
			@Override
			public void onResult(Result result) {
				handleResponse(result, successCallback, errorCallback);
			}
		});
	}

	@ReactMethod
	public void patch(final String url, final String body, final Callback successCallback, final Callback errorCallback) {
		RequestService.patch(url, body, getThirdPartyConnectorLibrary(), new ShcCallback() {
			@Override
			public void onResult(Result result) {
				handleResponse(result, successCallback, errorCallback);
			}
		});
	}

	private void sendEvent(String eventName, @Nullable WritableMap params) {
		this
			.reactContext
			.getJSModule(DeviceEventManagerModule.RCTDeviceEventEmitter.class)
			.emit(eventName, params);
	}

	private void handleResponse(final Result result, final Callback successCallback, final Callback errorCallback) {
		if (result.getCause() == null) {
			WritableMap params = Arguments.createMap();
			params.putInt("code", result.getCode());
			if (!result.getMessage().isEmpty()) {
				try {
					Object json = new JSONTokener(result.getMessage()).nextValue();
					if (json instanceof JSONObject) {
						params.putMap("body", JsonConverter.convertJsonToMap(new JSONObject(result.getMessage())));
					} else {
						params.putArray("body", JsonConverter.convertJsonToArray(new JSONArray(result.getMessage())));
					}
				} catch (JSONException e) {
					errorCallback.invoke(String.valueOf(result.getCode()), e.getMessage());
					return;
				}
			}
			if ((result.getCode() >= 200) && (result.getCode() < 300)) {
				successCallback.invoke(params);
			} else {
				errorCallback.invoke(params);
			}
		} else {
			errorCallback.invoke(String.valueOf(result.getCode()), result.getCause().getMessage());
		}
	}

	private void sendShcIdChangedEvent(final ShcIdChangeResult shcIdChangeResult) {
		Log.v(TAG, "SHC ID changed: " + shcIdChangeResult.name());
		WritableMap params = Arguments.createMap();
		params.putString("status", shcIdChangeResult.name());
		sendEvent(RNSHC_ID_CHANGE_RESULT, params);
	}
	
}