//
//  RNShcConnector.m
//  RNShcConnector
//
//  Created by Beata Brassai on 25/04/2018.
//  Copyright © 2018 Bosch TT. All rights reserved.
//

#import "RNShcConnector.h"
#import <React/RCTLog.h>
#import <ShcConRepresentation.h>
#import "RequestService.h"
#import <NSError+ShcConRepresentation.h>

@interface RNShcConnector ()
@property(nonatomic, strong) ShcConRepresentation* shcConRepresentation;
@property(nonatomic, strong) RequestService* requestService;
@end

NSString *const onShcConnected = @"RNSHC_CONNECTED";
NSString *const testEvent = @"RNSHC_TEST_EVENT";
NSString *const onShcConnectionLost = @"RNSHC_CONNECTION_LOST";
NSString *const onShcDisconnected = @"RNSHC_DISCONNECTED";
NSString *const onShcPairingStatus = @"RNSHC_PARING_STATUS";

@implementation RNShcConnector

RCT_EXPORT_MODULE()

- (NSArray<NSString *> *)supportedEvents
{
    return @[onShcConnected, testEvent, onShcConnectionLost, onShcDisconnected, onShcPairingStatus];
}

- (dispatch_queue_t)methodQueue
{
    return dispatch_get_main_queue();
}

RCT_EXPORT_METHOD(initModule)
{
    RCTLogInfo(@"initModule");
    self.shcConRepresentation = [[ShcConRepresentation alloc] init];
    self.requestService = [[RequestService alloc] initWithConRepresentation:self.shcConRepresentation notificationHandler:self];
}

RCT_EXPORT_METHOD(connect)
{
    // In case of the iOS side the connect function is out of scope
    RCTLogInfo(@"Connect");
}

RCT_EXPORT_METHOD(hasShcIdBeenProvided:(RCTPromiseResolveBlock)resolve reject:(RCTPromiseRejectBlock)reject)
{
    @try {
        [self.shcConRepresentation configureDefaultConRepresentationWithCompletionBlock:^(NSError * _Nullable error) {
            if (error != nil){
                // ShcConfigurationErrorDefaultConfigurationNotAvailable error code comes when the paired data is missing
                if ([error isConnectorErrorWithErrorCode:ShcConfigurationErrorDefaultConfigurationNotAvailable]) {
                    reject(@"SHC_ID_NOT_SET", @"The Shc Id has not been provided yet", error);
                } else {
                    reject(@"CONNECTION_ERROR", @"The configured SHC is not available", error);
                }
            } else {
                resolve(@"The SHC ID is already configured");
            }
        }];
    } @catch (NSException *ex) {
        [self reInitModule];
        reject(@"SHC_ID_NOT_SET", @"The configured SHC is not available", [NSError errorWithDomain:@"conlib" code:0 userInfo:ex.userInfo ]);
    }
}

RCT_EXPORT_METHOD(setShcId:(NSString *)id resolve:(RCTPromiseResolveBlock)resolve reject:(RCTPromiseRejectBlock)reject)
{
    RCTLogInfo(@"setShcId %@", id);
    [self.shcConRepresentation configureWithMACAddress:id completionBlock:^(NSError * _Nullable error) {
        if (error != nil)
        {
            reject(nil, @"setShcId is failed", error);
            return;
        }
        resolve(@[@"SHC ID is succsefully set"]);
    }];
}

RCT_EXPORT_METHOD(setShcIdAndLocalAddress:(NSString *)shcId address:(NSString *)address resolve:(RCTPromiseResolveBlock)resolve reject:(RCTPromiseRejectBlock)reject)
{
    [self.shcConRepresentation configureWithMACAddress:shcId ipAddress:address completionBlock:^(NSError * _Nullable error) {
        if (error != nil)
        {
            reject(nil, @"Set SHC ID and LocalAddress is failed", error);
            return;
        }
        resolve(@[@"SHC ID and IP address is succsefully set"]);
    }];
}

RCT_EXPORT_METHOD(checkPairStatus:(RCTPromiseResolveBlock)resolve reject:(RCTPromiseRejectBlock)reject)
{
    NSString *str;
    if (self.shcConRepresentation.connectivityState == ShcConnectivityCompatible)
    {
        str = [self stringFromState: self.shcConRepresentation.state];
    }
    else {
        str = [self stringFromConnectivityState: self.shcConRepresentation.connectivityState];
    }
    
    NSDictionary *dictionary = @{@"status" : str};
    RCTLogInfo(@"status %@", str);
    resolve(dictionary);
}

RCT_EXPORT_METHOD(pair:(NSString *)pairingId nonce:(NSString *)nonce  resolve:(RCTPromiseResolveBlock)resolve reject:(RCTPromiseRejectBlock)reject)
{
    RCTLogInfo(@"pairwith nonce");
    [self.requestService pairWithNonce:nonce pairingId:pairingId success:^(NSString* _Nullable response)
     {
         resolve(@[@"Pair successs"]);
     } failure:^(NSError * _Nonnull error, id  _Nullable rawResponse)
     {
         reject(nil, @"The pair with nonce has failed", error);
     }];
}

RCT_EXPORT_METHOD(get:(NSString *)url successCallback:(RCTResponseSenderBlock)successCallback errorCallback:(RCTResponseSenderBlock)errorCallback)
{
    [self.requestService sendGetByUrl:url successBlock:^(NSString * _Nullable response) {
        RCTLogInfo(@"Get request succeeded on URL: %@", url);
        successCallback(@[response]);
    } failureBlock:^(NSError * _Nonnull error, id  _Nullable rawResponse) {
        RCTLogInfo(@"Get request failed on URL %@.", url);
        errorCallback(@[[self createErrorResponse:error rawResponse:rawResponse]]);
    }];
}

RCT_EXPORT_METHOD(put:(NSString *)url body:(NSString *)body successCallback:(RCTResponseSenderBlock)successCallback errorCallback:(RCTResponseSenderBlock)errorCallback)
{
    [self.requestService sendPutByUrl:url body:body successBlock:^(NSString * _Nullable response) {
        RCTLogInfo(@"Put request succeeded on URL: %@", url);
        successCallback(@[response]);
    } failureBlock:^(NSError * _Nonnull error, id  _Nullable rawResponse) {
        RCTLogInfo(@"Put request failed on URL: %@.", url);
        errorCallback(@[[self createErrorResponse:error rawResponse:rawResponse]]);
    }];
}

RCT_EXPORT_METHOD(patch:(NSString *)url body:(NSString *)body successCallback:(RCTResponseSenderBlock)successCallback errorCallback:(RCTResponseSenderBlock)errorCallback)
{
    [self.requestService sendPatchByUrl:url body:body successBlock:^(NSString * _Nullable response) {
        RCTLogInfo(@"Patch request succeeded on URL %@", url);
        successCallback(@[response]);
    } failureBlock:^(NSError * _Nonnull error, id  _Nullable rawResponse) {
        RCTLogInfo(@"Patch request failed on URL: %@.", url);
        errorCallback(@[[self createErrorResponse:error rawResponse:rawResponse]]);
    }];
}

RCT_EXPORT_METHOD(clearPairData)
{
    [self.shcConRepresentation clearPairData];
}

RCT_EXPORT_METHOD(refreshConnection)
{
    [self.shcConRepresentation refreshConnection];
}

- (NSString*)stringFromState:(ShcConRepresentationState)state {
    switch (state) {
        case 0: return @"ShcConRepresentationNotConfigured";
        case 1: return @"ShcConRepresentationNotPaired";
        case 2: return @"ShcConRepresentationPaired";
        default: return @"Unknown";
    }
}

- (NSString*)stringFromConnectivityState:(ShcConnectivityState)state {
    switch (state) {
        case 0: return @"ShcConnectivityUnknown";
        case 1: return @"ShcConnectivityCompatible";
        case 2: return @"ShcConnectivitySHCOutdated";
        case 3: return @"ShcConnectivityClientOutdated";
        default: return @"Unknown";
    }
}

- (void)reInitModule
{
    RCTLogInfo(@"Clean up the stored data");
    [ShcConRepresentation cleanup];
    [self clearPairData];
    self.shcConRepresentation = nil;
    self.requestService = nil;
    [self initModule];
}

- (void)sendEvent:(NSString *)eventName body:(NSDictionary * _Nullable)body
{
    [self sendEventWithName:eventName body:body];
}

- (void)handleConnectedNotification:(NSString *)notification
{
    [self sendEvent:onShcConnected body:nil];
}

- (void)handleConnectLostNotification:(NSString *)notification
{
    [self sendEvent:onShcConnectionLost body:nil];
}

- (void)handleDisconnectedNotification:(NSString *)notification
{
    [self sendEvent:onShcDisconnected body:nil];
}

- (NSString*)createErrorResponse:(NSError* _Nonnull)error rawResponse:(id _Nullable)rawResponse
{
    if ([error isHTTPError]) {
        if (rawResponse != nil && [rawResponse isKindOfClass:NSData.class]) {
            NSString* body = [[NSString alloc] initWithData:rawResponse encoding:NSUTF8StringEncoding];
            NSMutableDictionary* errorJson = [[NSMutableDictionary alloc]init];
            [errorJson setValue:body forKey:@"body"];
            [errorJson setValue:@(error.code) forKey:@"code"];
            RCTLogInfo(@"RawResponse: %@, LocalizedError: %@", body, error.localizedDescription);
            NSError* dictionaryConversionError;
            NSData* jsonData = [NSJSONSerialization dataWithJSONObject:errorJson options:0 error:&dictionaryConversionError];
            if (dictionaryConversionError == nil) {
                NSString* jsonString = [[NSString alloc] initWithData:jsonData encoding:NSUTF8StringEncoding];
                return jsonString;
            } else {
                RCTLogInfo(@"Error during dictionary conversion: %@", dictionaryConversionError.localizedDescription);
                return error.localizedDescription;
            }
        } else {
            RCTLogInfo(@"LocalizedError: %@", error.localizedDescription);
            return error.localizedDescription;
        }
    } else {
        RCTLogInfo(@"LocalizedError: %@", error.localizedDescription);
        return error.localizedDescription;
    }
}

@end
