
package com.bosch.tt.privacypolicy.rn;

import android.util.Log;

import com.bosch.tt.privacypolicy.core.Country;
import com.bosch.tt.privacypolicy.core.Document;
import com.bosch.tt.privacypolicy.core.MetaDocument;
import com.bosch.tt.privacypolicy.core.TTPrivacyProvider;
import com.bosch.tt.privacypolicy.core.TTPrivacyProviderBuilder;
import com.bosch.tt.privacypolicy.rn.util.PrivacyPolicyOfflineProvider;
import com.bosch.tt.privacypolicy.rn.util.TermsOfUseProvider;
import com.bosch.tt.privacypolicy.ui.CountrySelector;
import com.bosch.tt.privacypolicy.ui.TTPrivacyComponent;
import com.bosch.tt.privacypolicy.valueprovider.PreviewValueProvider;
import com.bosch.tt.privacypolicy.valueprovider.ProductionValueProvider;
import com.bosch.tt.privacypolicy.valueprovider.ValueProvider;
import com.bosch.tt.privacypolicy.rn.assembler.CountryAssembler;
import com.bosch.tt.privacypolicy.rn.assembler.DocumentAssembler;
import com.bosch.tt.privacypolicy.rn.util.PrivacyPolicyEvent;
import com.facebook.react.bridge.Promise;
import com.facebook.react.bridge.ReactApplicationContext;
import com.facebook.react.bridge.ReactContextBaseJavaModule;
import com.facebook.react.bridge.ReactMethod;
import com.facebook.react.bridge.ReadableArray;
import com.facebook.react.bridge.ReadableMap;
import com.facebook.react.bridge.WritableArray;
import com.facebook.react.bridge.WritableMap;
import com.facebook.react.uimanager.IllegalViewOperationException;
import com.facebook.react.uimanager.NativeViewHierarchyManager;
import com.facebook.react.uimanager.UIBlock;
import com.facebook.react.uimanager.UIManagerModule;
import com.facebook.react.uimanager.events.RCTEventEmitter;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class RNPrivacyComponentModule extends ReactContextBaseJavaModule {

    private static final String TAG = RNPrivacyComponentModule.class.getSimpleName();;

    private final ReactApplicationContext reactContext;

    private CountryAssembler countryAssembler;
    private DocumentAssembler documentAssembler;

    private TTPrivacyProvider privacyProvider;
    private List<Country> supportedCountries;
    private Country selectedCountry;
    private CountrySelector.CountrySelectedListener countrySelectedListener;

    // Every TTPrivacyComponent has its own listener identified by its id
    private Map<Integer, TTPrivacyComponent.OnLoadFinishedListener> onDocumentLoadListeners;

    private TTPrivacyProvider termsOfUseProvider;

    public RNPrivacyComponentModule(final ReactApplicationContext reactContext) {
        super(reactContext);
        this.reactContext = reactContext;
        this.countryAssembler = new CountryAssembler(reactContext);
        this.documentAssembler = new DocumentAssembler();
        this.onDocumentLoadListeners = new HashMap<>();
        this.countrySelectedListener = new CountrySelector.CountrySelectedListener() {
            @Override
            public void onCountrySelected(CountrySelector countrySelector, Country country) {
                countrySelector.setSelectedCountry(country);
                reactContext.getJSModule(RCTEventEmitter.class).receiveEvent(countrySelector.getId(), PrivacyPolicyEvent.ON_COUNTRY_SELECTED.toString(), countryAssembler.createWritableDTO(country));
            }
        };
    }

    @Override
    public String getName() {
        return "RNPrivacyComponent";
    }

    @ReactMethod
    public void init(final String leanIX, final boolean productionMode) {
        final ValueProvider onlineProvider = productionMode ? new ProductionValueProvider(leanIX) : new PreviewValueProvider(leanIX);
        this.privacyProvider = new TTPrivacyProviderBuilder(this.reactContext, leanIX, new PrivacyPolicyOfflineProvider(this.reactContext)).onlineProvider(onlineProvider).build();
    }

    @ReactMethod
    public void loadPrivacyPolicy(final ReadableMap country, final int viewTag) {
        this.loadDocumentWithProvider(this.privacyProvider, country, viewTag);
    }

    @ReactMethod
    public void loadCountries(final int viewTag) {
        final UIManagerModule uiManager = this.reactContext.getNativeModule(UIManagerModule.class);
        uiManager.addUIBlock(new UIBlock() {
            @Override
            public void execute(NativeViewHierarchyManager nativeViewHierarchyManager) {
                try {
                    final CountrySelector countrySelector = (CountrySelector) nativeViewHierarchyManager.resolveView(viewTag);
                    countrySelector.loadCountries(privacyProvider, supportedCountries, selectedCountry, countrySelectedListener);
                    Log.d(TAG, "loadCountries method got called");
                } catch (IllegalViewOperationException e) {
                    Log.e(TAG, e.getMessage());
                }
            }
        });
    }

    @ReactMethod
    public void setSupportedCountries(final ReadableArray countries, final int viewTag) {
        this.supportedCountries = countryAssembler.createModelArray(countries);
        final UIManagerModule uiManager = this.reactContext.getNativeModule(UIManagerModule.class);
        uiManager.addUIBlock(new UIBlock() {
            @Override
            public void execute(NativeViewHierarchyManager nativeViewHierarchyManager) {
                try {
                    final CountrySelector countrySelector = (CountrySelector) nativeViewHierarchyManager.resolveView(viewTag);
                    countrySelector.setCountrySelectedListener(countrySelectedListener);
                    countrySelector.setCountries(supportedCountries);
                    Log.d(TAG, "setSupportedCountries method got called with " + supportedCountries.size() + " countries.");
                } catch (IllegalViewOperationException e) {
                    Log.e(TAG, e.getMessage());
                }
            }
        });
    }

    @ReactMethod
    public void setSelectedCountry(final ReadableMap country, final int viewTag) {
        this.selectedCountry = countryAssembler.createModel(country);
        final UIManagerModule uiManager = this.reactContext.getNativeModule(UIManagerModule.class);
        uiManager.addUIBlock(new UIBlock() {
            @Override
            public void execute(NativeViewHierarchyManager nativeViewHierarchyManager) {
                try {
                    final CountrySelector countrySelector = (CountrySelector) nativeViewHierarchyManager.resolveView(viewTag);
                    countrySelector.setSelectedCountry(selectedCountry);
                    Log.d(TAG, "setSelectedCountry method got called with " + selectedCountry.toString());
                } catch (IllegalViewOperationException e) {
                    Log.e(TAG, e.getMessage());
                }
            }
        });
    }

    @ReactMethod
    public void accept(final String documentId) {
        this.privacyProvider.accept(documentId);
    }

    @ReactMethod
    public void isAccepted(final String documentId, final Promise promise) {
        promise.resolve(this.privacyProvider.isAccepted(documentId));
    }

    @ReactMethod
    public void getAvailableDocuments(final Promise promise) {
        this.getAvailableDocumentsWithProvider(this.privacyProvider, promise);
    }

    @ReactMethod
    public void getLastCheckedTimestamp(final String country, final String language, final Promise promise) {
        promise.resolve(this.privacyProvider.getLastCheckedTimestamp(country, language).toString());
    }

    @ReactMethod
    public void getTimestampOfConsent(final String documentId, final Promise promise) {
        promise.resolve(this.privacyProvider.getTimestampOfConsent(documentId).toString());
    }

    @ReactMethod
    public void getAvailableCountries(final ReadableArray countriesFromApp, final Promise promise) {
        final List<Country> countries = this.countryAssembler.createModelArray(countriesFromApp);
        final WritableArray countriesDTO = this.countryAssembler.createWritableDTOArray(this.privacyProvider.getAvailableCountries(countries));
        promise.resolve(countriesDTO);
    }

    @ReactMethod
    public void getDocument(final String country, final String language, final Promise promise) {
        this.getDocumentWithProvider(this.privacyProvider, country, language, promise);
    }

    @ReactMethod
    public void loadTermsOfUse(final ReadableMap country, final int viewTag) {
        this.loadDocumentWithProvider(getTermsOfUseProvider(), country, viewTag);
    }

    @ReactMethod
    public void getTermsOfUseDocument(final String country, final String language, final Promise promise) {
        this.getDocumentWithProvider(getTermsOfUseProvider(), country, language, promise);
    }

    @ReactMethod
    public void getAvailableTermsOfUseDocuments(final Promise promise) {
        this.getAvailableDocumentsWithProvider(getTermsOfUseProvider(), promise);
    }

    private TTPrivacyComponent.OnLoadFinishedListener getOnDocumentLoadListener(final TTPrivacyComponent privacyComponent) {
        if (!this.onDocumentLoadListeners.containsKey(privacyComponent.getId())) {
            final TTPrivacyComponent.OnLoadFinishedListener onLoadFinishedListener = new TTPrivacyComponent.OnLoadFinishedListener() {
                @Override
                public void onLoadFinished(Document document) {
                    reactContext.getJSModule(RCTEventEmitter.class).receiveEvent(privacyComponent.getId(), PrivacyPolicyEvent.ON_DOCUMENT_LOADED.toString(), documentAssembler.createWritableDTO(document));
                }
            };
            this.onDocumentLoadListeners.put(privacyComponent.getId(), onLoadFinishedListener);
        }
        return this.onDocumentLoadListeners.get(privacyComponent.getId());
    }

    private TTPrivacyProvider getTermsOfUseProvider() {
        if (this.termsOfUseProvider == null) {
            final ValueProvider valueProvider = new TermsOfUseProvider(this.reactContext);
            this.termsOfUseProvider = new TTPrivacyProviderBuilder(this.reactContext, null, valueProvider).onlineProvider(valueProvider).build();
        }
        return termsOfUseProvider;
    }

    private void getDocumentWithProvider(final TTPrivacyProvider provider, final String country, final String language, final Promise promise) {
        final Document document = provider.getDocument(country, language);
        if (document != null) {
            final WritableMap documentDTO = this.documentAssembler.createWritableDTO(document);
            promise.resolve(documentDTO);
        } else {
            Log.w(TAG, "No document found for country: " + country + ", language: " + language);
            promise.reject("404", "No document found!");
        }
    }

    private void getAvailableDocumentsWithProvider(final TTPrivacyProvider provider, final Promise promise) {
        final List<MetaDocument> metaDocuments = provider.getAvailableDocuments();
        if (metaDocuments != null) {
            final WritableArray readableArray = this.documentAssembler.createWritableDTOArray(metaDocuments);
            promise.resolve(readableArray);
        } else {
            Log.w(TAG, "No available documents found!");
            promise.reject("404", "No available documents!");
        }
    }

    private void loadDocumentWithProvider(final TTPrivacyProvider provider, final ReadableMap country, final int viewTag) {
        final UIManagerModule uiManager = this.reactContext.getNativeModule(UIManagerModule.class);
        uiManager.addUIBlock(new UIBlock() {
            @Override
            public void execute(NativeViewHierarchyManager nativeViewHierarchyManager) {
                try {
                    final Country selectedCountry = countryAssembler.createModel(country);
                    final TTPrivacyComponent privacyComponent = (TTPrivacyComponent) nativeViewHierarchyManager.resolveView(viewTag);
                    privacyComponent.loadDocument(provider, selectedCountry.getCode(), selectedCountry.getLanguage(), getOnDocumentLoadListener(privacyComponent));
                    Log.d(TAG, "loadDocument got called with: " + selectedCountry.toString());
                } catch (IllegalViewOperationException e) {
                    Log.e(TAG, e.getMessage());
                }
            }
        });
    }

}