
import { Component } from 'react';
import { ViewProperties } from 'react-native';

declare module "react-native-privacy-component" {

    export interface Country {
        flagRes?: string;
        name: string;
        code: string;
        language: string;
    }

    export interface MetaDocument {
        id: string;
        country: string;
        language: string;
        validFrom: string;
        createdOn: string;
    }

    export enum DocumentType {
        PrivacyPolicy = 'PrivacyPolicy',
        TermsOfUse = 'TermsOfUse'
    }

    export enum ActivityIndicatorStyle {
        WhiteLarge = 'UIActivityIndicatorViewStyleWhiteLarge',
        White = 'UIActivityIndicatorViewStyleWhite',
        Gray = 'UIActivityIndicatorViewStyleGray'
    }

    export default class RNPrivacyComponent {
        /**
         * Initializes the underlying TTPrivacyProvider class.
         * @param leanIX - project's lean id
         * @param productionMode - use the production API or the preview
         */
        static init(leanIX: string, productionMode: boolean): void;
        
        /**
         * Accepts a given document.
         * @param documentId - the document's id
         */
        static accept(documentId: string): void;
        
        /**
         * Checks the status of a document.
         * @param documentId - the document's id
         * @returns Promise with a boolean value
         */
        static isAccepted(documentId: string): Promise<boolean>;

        /**
         * Retrieves the available meta documents.
         * @returns - Promise object with a MetaDocument array
         */
        static getAvailableDocuments(): Promise<MetaDocument[]>;

        /**
         * Returns with the timestamp of the last online check.
         * @param country - country code
         * @param language - language code
         * @returns Promise object with a Date as string
         */
        static getLastCheckedTimestamp(country: string, language: string): Promise<string>;

        /**
         * Retrieves the accept date of a document.
         * @param documentId - the document's id
         * @returns Promise object with a Date as string
         */
        static getTimestampOfConsent(documentId: string): Promise<string>;

        /**
         * Returns the common items from the online and offline countries. "Online countries" means that it available on the privacy portal. 
         * "Offline countries" means that the app stores the privacy texts for this countries by default.
         * @param countriesFromApp - list of the supported offline countries
         * @returns Promise object with the common items from the online and offline countries
         */
        static getAvailableCountries(countriesFromApp: Country[]): Promise<Country[]>;

        /**
         * Returns a document based on country and language. If the server is unreachable it uses the offline value provider.
         * @param country - country code
         * @param language - language code
         * @returns Promise object containing the requested document
         */
        static getDocument(country: string, language: string): Promise<MetaDocument>;

        /**
         * Retrieves the available terms of use meta documents.
         * @returns - Promise object with a MetaDocument array
         */
        static getAvailableTermsOfUseDocuments(): Promise<MetaDocument[]>;

        /**
         * Returns a terms of use document based on country and language.
         * @param country - country code
         * @param language - language code
         * @returns Promise object containing the requested document
         */
        static getTermsOfUseDocument(country: string, language: string): Promise<MetaDocument>;
    }

    interface TTPrivacyComponentProps {
        /**
         * Shows the privacy policy based on this country.
         */
        selectedCountry?: Country;
        /**
         * Applies this style to the HTML content.
         * e.g.: "body {color: #900; text-align:justify; font-size:80%;}"
         */
        css?: string;
        /**
         * Returns the currently loaded document's meta data when the requested privacy policy appears.
         * @param document - the requested document's meta data
         */
        onDocumentLoaded?(document: MetaDocument): void;
        /**
         * Gets called when the ScrollView reaches its bottom.
         */
        onScrollEnded?(): void;
        /**
         * Type of the displayed document. Default is PrivacyPolicy.
         */
        type?: DocumentType;
        /**
         * iOS only
         * Sets the style of the activity indicator on TTPrivacyComponent
         */
        activityIndicatorStyle?: ActivityIndicatorStyle
    }

    interface CountrySelectorProps {
        /**
         * Which countries are supported by the application.
         */
        countries?: Country[];
        /**
         * Relevant only on Android. Shows the common countries from the available documents and
         * the provided countries from prop.
         */
        loadCountries?: boolean;
        /**
         * Gets called in case the user taps on one of the displayed countries.
         * @param country on which the user tapped
         */
        onCountrySelected?(country: Country): void;
    }

    export class TTPrivacyComponent extends Component<ViewProperties & TTPrivacyComponentProps> {}
    export class CountrySelector extends Component<ViewProperties & CountrySelectorProps> {}
}