//
//  RNPrivacyComponent.m
//  RNPrivacyComponent
//
//  Created by István Bege on 26/10/2018.
//  Copyright (C) 2018, Bosch Thermotechnik GmbH. All rights reserved.
//

#import <React/RCTComponent.h>

#import <TTPrivacyComponent/TTPMetaDocument.h>

#import "RNPrivacyComponent.h"
#import "PrivacyManager.h"
#import "Assembler.h"

@interface RNPrivacyComponent ()

@property (readonly) PrivacyManager* privacyManager;
@property (readonly) NSDateFormatter* dateFormatter;

@end

@implementation RNPrivacyComponent

- (dispatch_queue_t)methodQueue
{
    return dispatch_get_main_queue();
}

RCT_EXPORT_MODULE()

RCT_EXPORT_METHOD(init:(NSString*)leanIX productionMode:(BOOL)productionMode)
{
    _privacyManager = [PrivacyManager getInstance];
    [_privacyManager initProviderWithLeanIX:leanIX productionMode:productionMode];
    _dateFormatter = [[NSDateFormatter alloc] init];
    [_dateFormatter setDateFormat:@"yyyy-MM-dd'T'HH:mm:ss"];
}

# pragma mark Country specific methods

RCT_EXPORT_METHOD(setSupportedCountries:(NSArray<NSDictionary*>*)supportedCountries reactTag:(nonnull NSNumber *)reactTag)
{
    [_privacyManager setSupportedCountries:[Assembler createCountryArray:supportedCountries]];
}

RCT_EXPORT_METHOD(setSelectedCountry:(NSDictionary*)rawCountry reactTag:(nonnull NSNumber *)reactTag)
{
    [_privacyManager setSelectedCountry:[Assembler createCountry:rawCountry]];
}

RCT_EXPORT_METHOD(getAvailableCountries:(NSArray<NSDictionary*>*) countriesFromApp resolver:(RCTPromiseResolveBlock)resolve rejecter:(RCTPromiseRejectBlock)reject)
{
    NSArray<TTPCountry*>* availableCountries = [[_privacyManager privacyProvider] getAvailableCountries:[Assembler createCountryArray:countriesFromApp]];
    resolve([Assembler createCountryDTOArray:availableCountries]);
}

# pragma mark Privacy Policy specific methods

RCT_EXPORT_METHOD(accept:(NSString*)documentId)
{
    [[_privacyManager privacyProvider] accept:documentId];
}

RCT_EXPORT_METHOD(isAccepted:(NSString*)documentId resolver:(RCTPromiseResolveBlock)resolve rejecter:(RCTPromiseRejectBlock)reject)
{
    BOOL result = [[_privacyManager privacyProvider] isAccepted:documentId];
    resolve([NSNumber numberWithBool:result]);
}

RCT_EXPORT_METHOD(getAvailableDocuments:(RCTPromiseResolveBlock)resolve rejecter:(RCTPromiseRejectBlock)reject)
{
    [self getAvailableDocumentWithProvider:[_privacyManager privacyProvider] resolver:resolve rejecter:reject];
}

RCT_EXPORT_METHOD(getLastCheckedTimestamp:(NSString*)country language:(NSString*)language resolver:(RCTPromiseResolveBlock)resolve rejecter:(RCTPromiseRejectBlock)reject)
{
    NSDate* result = [[_privacyManager privacyProvider] lastCheckedTimestampWithCountry:country language:language];
    if (result != nil) {
        resolve([[NSString alloc] initWithString:[_dateFormatter stringFromDate:result]]);
    } else {
        reject(@"404", @"The document was not checked on the required country and language.", nil);
    }
}

RCT_EXPORT_METHOD(getTimestampOfConsent:(NSString*)documentId resolver:(RCTPromiseResolveBlock)resolve rejecter:(RCTPromiseRejectBlock)reject)
{
    NSDate* result = [[_privacyManager privacyProvider] getTimestampOfConsent:documentId];
    if (result != nil) {
        resolve([[NSString alloc] initWithString:[_dateFormatter stringFromDate:result]]);
    } else {
        reject(@"404", @"The provided document was not accepted.", nil);
    }
}

RCT_EXPORT_METHOD(getDocument:(NSString*)country language:(NSString*)language resolver:(RCTPromiseResolveBlock)resolve rejecter:(RCTPromiseRejectBlock)reject)
{
    [self getDocumentWithProvider:[_privacyManager privacyProvider] country:country language:language resolver:resolve rejecter:reject];
}

# pragma mark Terms of Use specific methods

RCT_EXPORT_METHOD(getAvailableTermsOfUseDocuments:(RCTPromiseResolveBlock)resolve rejecter:(RCTPromiseRejectBlock)reject)
{
    [self getAvailableDocumentWithProvider:[_privacyManager getTermsOfUseProvider] resolver:resolve rejecter:reject];
}

RCT_EXPORT_METHOD(getTermsOfUseDocument:(NSString*)country language:(NSString*)language resolver:(RCTPromiseResolveBlock)resolve rejecter:(RCTPromiseRejectBlock)reject)
{
    [self getDocumentWithProvider:[_privacyManager getTermsOfUseProvider] country:country language:language resolver:resolve rejecter:reject];
}

# pragma mark private methods

- (void)getDocumentWithProvider:(TTPPrivacyProvider*)provider country:(NSString*)country language:(NSString*)language resolver:(RCTPromiseResolveBlock)resolve rejecter:(RCTPromiseRejectBlock)reject
{
    TTPMetaDocument* document = [provider getDocumentWithCountry:country language:language];
    if (document != nil) {
        resolve([Assembler createDocumentDTO:document]);
    } else {
        reject(@"404", @"No document found!", nil);
    }
}

- (void)getAvailableDocumentWithProvider:(TTPPrivacyProvider*)provider resolver:(RCTPromiseResolveBlock)resolve rejecter:(RCTPromiseRejectBlock)reject
{
    NSArray<TTPMetaDocument*>* documents = [provider getAvailableDocuments];
    if (documents != nil) {
        resolve([Assembler createDocumentDTOArray:documents]);
    } else {
        reject(@"404", @"No available documents!", nil);
    }
}

@end
